-------------------------------------------------------------------------------
-- Description: A NoC 2d-Mesh composed of NxM routers
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.noc_pack.all;

entity noc_mesh2d is

  generic (
    x   : integer := 4;
    y   : integer := 4);
  port (
    clk : in std_logic;                 -- clock
    rst : in std_logic);                -- reset

end noc_mesh2d;

architecture noc_mesh2d_arch of noc_mesh2d is
  
  -- noc data and ctrl array to connect routers in a 2d mesh
  type nocph_data_2darray_type is array (integer range <>,integer range <>) of nocph_data_type;
  type nocph_ctrl_2darray_type is array (integer range <>,integer range <>) of nocph_ctrl_type;
  
  -- horizontal west and east router interconnect
  signal west_data     : nocph_data_2darray_type(0  to x  , 0 to y-1);
  signal west_ctrl     : nocph_ctrl_2darray_type(0  to x  , 0 to y-1);
  signal east_data     : nocph_data_2darray_type(-1 to x-1, 0 to y-1);
  signal east_ctrl     : nocph_ctrl_2darray_type(-1 to x-1, 0 to y-1);
  -- vertical north and south router interconnect
  signal south_data    : nocph_data_2darray_type(0  to x-1, 0 to y  );
  signal south_ctrl    : nocph_ctrl_2darray_type(0  to x-1, 0 to y  );
  signal north_data    : nocph_data_2darray_type(0  to x-1,-1 to y-1);
  signal north_ctrl    : nocph_ctrl_2darray_type(0  to x-1,-1 to y-1);
  -- resource connections to/from switches
  signal res_tx_data   : nocph_data_2darray_type(0  to x-1, 0 to y-1);
  signal res_tx_ctrl   : nocph_ctrl_2darray_type(0  to x-1, 0 to y-1);
  signal res_rx_data   : nocph_data_2darray_type(0  to x-1, 0 to y-1);
  signal res_rx_ctrl   : nocph_ctrl_2darray_type(0  to x-1, 0 to y-1);
  
begin  -- noc_mesh2d_arch

  -----------------------------------------------------------------------------
  -- build the 2d mesh
  -----------------------------------------------------------------------------
  noc_x_dimension : for i in 0 to x-1 generate  -- x-Coordinate
    noc_y_dimension : for j in 0 to y-1 generate  -- y-Coordinate

      -----------------------------------------------------------------------------
      -- instance of the router
      -----------------------------------------------------------------------------
      DUT : noc_router2d
        generic map (
          self_adress => std_logic_vector(to_unsigned(i,2)) &  std_logic_vector(to_unsigned(j,2)))
        port map (
          clk           =>  clk,
          rst           =>  rst,
          south_tx_data =>  south_data(i  ,j),
          south_tx_ctrl =>  south_ctrl(i  ,j),
          south_rx_data =>  north_data(i  ,j-1),
          south_rx_ctrl =>  north_ctrl(i  ,j-1),
          north_tx_data =>  north_data(i  ,j),
          north_tx_ctrl =>  north_ctrl(i  ,j),
          north_rx_data =>  south_data(i  ,j+1),
          north_rx_ctrl =>  south_ctrl(i  ,j+1),
          west_tx_data  =>   west_data(i  ,j),
          west_tx_ctrl  =>   west_ctrl(i  ,j),
          west_rx_data  =>   east_data(i-1,j),
          west_rx_ctrl  =>   east_ctrl(i-1,j),
          east_tx_data  =>   east_data(i  ,j),
          east_tx_ctrl  =>   east_ctrl(i  ,j),
          east_rx_data  =>   west_data(i+1,j),
          east_rx_ctrl  =>   west_ctrl(i+1,j),
          res_tx_data   => res_tx_data(i  ,j),
          res_tx_ctrl   => res_tx_ctrl(i  ,j),
          res_rx_data   => res_rx_data(i  ,j),
          res_rx_ctrl   => res_rx_ctrl(i  ,j));
      
    end generate noc_y_dimension;
  end generate noc_x_dimension;


end noc_mesh2d_arch;

