-------------------------------------------------------------------------------
-- Description: Constants, Types, Functions related to the NoC
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

package noc_pack is
  
  -----------------------------
  --   4   4    4    4    16  = 32
  -- |DST|SRC|PRIO|TYPE|UNUSED|
  -----------------------------
  -- Flit-Type
  subtype nocph_flit_type is std_logic_vector (31 downto 0);
  
  constant NOC_ZEROFLIT : nocph_flit_type := (others => '0');  
  constant NOC_ZERO_BE  : std_logic_vector(1 downto 0) := (others => '0');

  -- interface type between adjacent routers
  type nocph_data_type is record
                            data : nocph_flit_type;  -- Data (one flit)
                            req  : std_logic;        -- Request
                            sof  : std_logic;        -- Start of Frame
                            eof  : std_logic;        -- End of Frame
                            -- Byte-enable Signale
                            be   : std_logic_vector(1 downto 0);
                        end record;

  -- interface used within a router
  type nocdl_data_type is record
                           data : nocph_flit_type;
                           req  : std_logic;
                       end record;
  
  -- acknowledge/backpressure signal between adjacent switches
  type nocph_ctrl_type is record
                            ack : std_logic;
                        end record;
  
  -- acknowledge/backpressure signal within a switches
  type nocdl_ctrl_type is record
                           ack : std_logic;
                       end record;

  constant nocph_data_type_default : nocph_data_type := (NOC_ZEROFLIT, '0', '0', '0', NOC_ZERO_BE);
  constant nocdl_data_type_default : nocdl_data_type := (NOC_ZEROFLIT, '0');
  constant nocph_ctrl_type_default : nocph_ctrl_type := (ack => '0');
  constant nocdl_ctrl_type_default : nocdl_ctrl_type := (ack => '0');

  -- some address parts
  subtype noc_xadresspart is integer range 4-1   downto 4/2;
  subtype noc_yadresspart is integer range 4/2-1 downto 0;
  -- distinct parts in the headerflit
  subtype noc_adresspart  is integer range 32-1 downto 32-(2*4);
  subtype noc_dstpart     is integer range 32-1 downto 32-4;
  subtype noc_xdstpart    is integer range 32-1 downto 32-4/2;
  subtype noc_ydstpart    is integer range 32-4/2-1 downto 32-4;
  subtype noc_srcpart     is integer range 32-(4)-1 downto 32-(2*4);
  subtype noc_xsrcpart    is integer range 32-(4)-1 downto 32-(3*4/2);
  subtype noc_ysrcpart    is integer range 32-(3*4/2)-1 downto 32-(2*4);
  subtype noc_priopart    is integer range 32-(2*4)-1 downto 32-(2*4)-4;
  subtype noc_typepart    is integer range 32-(2*4)-4-1 downto 32-(2*4)-4-4;
  subtype noc_rsvdpart    is integer range 32-(2*4)-4-4-1 downto 0;

  function get_single_dim_target (
    own  : std_logic_vector(4/2-1 downto 0);
    dest : std_logic_vector(4/2-1 downto 0))
    return std_logic_vector;

  function get_target (
    own  : std_logic_vector(4-1 downto 0);
    dest : std_logic_vector(4-1 downto 0))
    return std_logic_vector;

  procedure grant_request (
    constant port_number : in  std_logic_vector (1 downto 0);
    variable state_out   : out std_logic_vector (1 downto 0);
    variable free        : out std_logic;
    variable ack_ar      : out std_logic);

  procedure grant_request5ports (
    constant port_number : in  std_logic_vector (2 downto 0);
    variable state_out   : out std_logic_vector (2 downto 0);
    variable free        : out std_logic;
    variable ack_ar      : out std_logic);

    component nocdl_fifo
    port (
      rst        : in  std_logic;
      clk        : in  std_logic;
      write_data : in  nocph_data_type;
      write_ctrl : out nocph_ctrl_type;
      read_data  : out nocph_data_type;
      read_ctrl  : in  nocph_ctrl_type);
  end component;

  component noc_westport2d
    generic (
      self_adress    : std_logic_vector (3 downto 0));
    port (
      clk            : in  std_logic;
      rst            : in  std_logic;
      data_out       : out nocph_data_type;
      ack_tx         : in  nocph_ctrl_type;
      data_in        : in  nocph_data_type;
      ack_rx         : out nocph_ctrl_type;
      free           : out std_logic;
      data_in_intern : out nocph_data_type;
      ack_tx_intern  : out nocph_ctrl_type;
      state_out      : in  std_logic_vector (1 downto 0);
      req_ar_east    : out std_logic;
      req_ar_north   : out std_logic;
      req_ar_south   : out std_logic;
      req_ar_res     : out std_logic;
      ack_ar_east    : in  nocdl_ctrl_type;
      ack_ar_north   : in  nocdl_ctrl_type;
      ack_ar_south   : in  nocdl_ctrl_type;
      ack_ar_res     : in  nocdl_ctrl_type;
      data_in_east   : in  nocph_data_type;
      data_in_res    : in  nocph_data_type;
      ack_tx_east    : in  nocph_ctrl_type;
      ack_tx_north   : in  nocph_ctrl_type;
      ack_tx_south   : in  nocph_ctrl_type;
      ack_tx_res     : in  nocph_ctrl_type);
  end component;

  component noc_eastport2d
    generic (
      self_adress    : std_logic_vector (3 downto 0));
    port (
      clk            : in  std_logic;
      rst            : in  std_logic;
      data_out       : out nocph_data_type;
      ack_tx         : in  nocph_ctrl_type;
      data_in        : in  nocph_data_type;
      ack_rx         : out nocph_ctrl_type;
      free           : out std_logic;
      data_in_intern : out nocph_data_type;
      ack_tx_intern  : out nocph_ctrl_type;
      state_out      : in  std_logic_vector (1 downto 0);
      req_ar_west    : out std_logic;
      req_ar_north   : out std_logic;
      req_ar_south   : out std_logic;
      req_ar_res     : out std_logic;
      ack_ar_west    : in  nocdl_ctrl_type;
      ack_ar_north   : in  nocdl_ctrl_type;
      ack_ar_south   : in  nocdl_ctrl_type;
      ack_ar_res     : in  nocdl_ctrl_type;
      data_in_west   : in  nocph_data_type;
      data_in_res    : in  nocph_data_type;
      ack_tx_west    : in  nocph_ctrl_type;
      ack_tx_north   : in  nocph_ctrl_type;
      ack_tx_south   : in  nocph_ctrl_type;
      ack_tx_res     : in  nocph_ctrl_type);
  end component;

  component noc_northport2d
    generic (
      self_adress    : std_logic_vector (3 downto 0));
    port (
      clk            : in  std_logic;
      rst            : in  std_logic;
      data_out       : out nocph_data_type;
      ack_tx         : in  nocph_ctrl_type;
      data_in        : in  nocph_data_type;
      ack_rx         : out nocph_ctrl_type;
      free           : out std_logic;
      data_in_intern : out nocph_data_type;
      ack_tx_intern  : out nocph_ctrl_type;
      state_out      : in  std_logic_vector (2 downto 0);
      req_ar_south   : out std_logic;
      req_ar_res     : out std_logic;
      ack_ar_south   : in  nocdl_ctrl_type;
      ack_ar_res     : in  nocdl_ctrl_type;
      data_in_west   : in  nocph_data_type;
      data_in_east   : in  nocph_data_type;
      data_in_south  : in  nocph_data_type;
      data_in_res    : in  nocph_data_type;
      ack_tx_south   : in  nocph_ctrl_type;
      ack_tx_res     : in  nocph_ctrl_type);
  end component;

  component noc_southport2d
    generic (
      self_adress     : std_logic_vector (3 downto 0));
    port (
      clk            : in  std_logic;
      rst            : in  std_logic;
      data_out       : out nocph_data_type;
      ack_tx         : in  nocph_ctrl_type;
      data_in        : in  nocph_data_type;
      ack_rx         : out nocph_ctrl_type;
      free           : out std_logic;
      data_in_intern : out nocph_data_type;
      ack_tx_intern  : out nocph_ctrl_type;
      state_out      : in  std_logic_vector (2 downto 0);
      req_ar_north   : out std_logic;
      req_ar_res     : out std_logic;
      ack_ar_north   : in  nocdl_ctrl_type;
      ack_ar_res     : in  nocdl_ctrl_type;
      data_in_west   : in  nocph_data_type;
      data_in_east   : in  nocph_data_type;
      data_in_north  : in  nocph_data_type;
      data_in_res    : in  nocph_data_type;
      ack_tx_north   : in  nocph_ctrl_type;
      ack_tx_res     : in  nocph_ctrl_type);
  end component;

  component noc_resport2d
    generic (
      self_adress    : std_logic_vector (3 downto 0));
    port (
      clk            : in  std_logic;
      rst            : in  std_logic;
      data_out       : out nocph_data_type;
      ack_tx         : in  nocph_ctrl_type;
      data_in        : in  nocph_data_type;
      ack_rx         : out nocph_ctrl_type;
      free           : out std_logic;
      data_in_intern : out nocph_data_type;
      ack_tx_intern  : out nocph_ctrl_type;
      state_out      : in  std_logic_vector (2 downto 0);
      req_ar_west    : out std_logic;
      req_ar_east    : out std_logic;
      req_ar_north   : out std_logic;
      req_ar_south   : out std_logic;
      ack_ar_west    : in  nocdl_ctrl_type;
      ack_ar_east    : in  nocdl_ctrl_type;
      ack_ar_north   : in  nocdl_ctrl_type;
      ack_ar_south   : in  nocdl_ctrl_type;
      data_in_west   : in  nocph_data_type;
      data_in_east   : in  nocph_data_type;
      data_in_north  : in  nocph_data_type;
      data_in_south  : in  nocph_data_type;
      ack_tx_west    : in  nocph_ctrl_type;
      ack_tx_east    : in  nocph_ctrl_type;
      ack_tx_north   : in  nocph_ctrl_type;
      ack_tx_south   : in  nocph_ctrl_type);
  end component;

  component noc_router2d
    generic (
      self_adress    : std_logic_vector (3 downto 0));
    port (
      clk           : in  std_logic;
      rst           : in  std_logic;
      south_tx_data : out nocph_data_type;
      south_tx_ctrl : in  nocph_ctrl_type;
      south_rx_data : in  nocph_data_type;
      south_rx_ctrl : out nocph_ctrl_type;
      north_tx_data : out nocph_data_type;
      north_tx_ctrl : in  nocph_ctrl_type;
      north_rx_data : in  nocph_data_type;
      north_rx_ctrl : out nocph_ctrl_type;
      west_tx_data  : out nocph_data_type;
      west_tx_ctrl  : in  nocph_ctrl_type;
      west_rx_data  : in  nocph_data_type;
      west_rx_ctrl  : out nocph_ctrl_type;
      east_tx_data  : out nocph_data_type;
      east_tx_ctrl  : in  nocph_ctrl_type;
      east_rx_data  : in  nocph_data_type;
      east_rx_ctrl  : out nocph_ctrl_type;
      res_tx_data   : out nocph_data_type;
      res_tx_ctrl   : in  nocph_ctrl_type;
      res_rx_data   : in  nocph_data_type;
      res_rx_ctrl   : out nocph_ctrl_type);
  end component;
  
end noc_pack;

package body noc_pack is

  function get_single_dim_target(
    own  : std_logic_vector(4/2-1 downto 0);
    dest : std_logic_vector(4/2-1 downto 0))
    return std_logic_vector is
   variable target : std_logic_vector (1 downto 0);
  begin  -- get_ytarget
    if own = dest
    then target := "10"; end if;
    if own < dest
    then target := "01"; end if;
    if own > dest
    then target := "11"; end if;
    return target;
  end get_single_dim_target;

  function get_target (
    own  : std_logic_vector(4-1 downto 0);
    dest : std_logic_vector(4-1 downto 0))
    return std_logic_vector is
  begin
    if own = dest then
      return "101";  
    else
      if own(4-1 downto 4/2) > dest(4-1 downto 4/2) then
        return "001";
      else  
        if own(4-1 downto 4/2) < dest(4-1 downto 4/2) then
          return "010";
        else
          if own(4/2-1 downto 0) > dest(4/2-1 downto 0) then
            return "100";
          else
            if own(4/2-1 downto 0) < dest(4/2-1 downto 0) then
              return "011";
            else
              return "000";
            end if;
          end if;
        end if;
      end if;
    end if;
  end get_target;

  procedure grant_request (
    constant port_number : in  std_logic_vector (1 downto 0);
    variable state_out   : out std_logic_vector (1 downto 0);
    variable free        : out std_logic;
    variable ack_ar      : out std_logic) is
  begin
    state_out := port_number;
    free      := '0';
    ack_ar    := '1';            
  end grant_request;

  procedure grant_request5ports (
    constant port_number : in  std_logic_vector (2 downto 0);
    variable state_out   : out std_logic_vector (2 downto 0);
    variable free        : out std_logic;
    variable ack_ar      : out std_logic) is
  begin
    state_out := port_number;
    free      := '0';
    ack_ar    := '1';            
  end grant_request5ports;
 
end noc_pack;

