-------------------------------------------------------------------------------
-- Description: Res-Port
--
--
-- 
-------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use work.noc_pack.all;

entity noc_resport2d is

  generic (
    SELF_ADRESS    : std_logic_vector (3 downto 0) := "0101");
  
  port (
    -- Standardsignale
    clk            : in  std_logic;               -- takt
    rst            : in  std_logic;               -- Reset

    -- I/O-Signale nach/von extern
    data_out       : out nocph_data_type;         -- ausgehendes Record aus Daten und Request
    ack_tx         : in  nocph_ctrl_type;         -- eingehendes Acknowledge fr zu sendende Daten
    data_in        : in  nocph_data_type;         -- eingehendes Record aus Daten und Request
    ack_rx         : out nocph_ctrl_type;         -- ausgehendes Acknowledge fr empfangene Daten

    -- I/O-Signale nach/von intern    
    free           : out std_logic;               -- Statusanzeige des Ports   
    data_in_intern : out nocph_data_type;         -- Daten+Request nach Intern
    ack_tx_intern  : out nocph_ctrl_type;
    state_out      : in  std_logic_vector (2 downto 0);                -- steuert data_in-Mux
     -- Ausgnge demux_req
    req_ar_east    : out std_logic;               -- Request-Signal fr Arbiter-Logik (East-Out = X-Plus-Out)
    req_ar_north   : out std_logic;               -- Request-Signal fr Arbiter-Logik (North-Out = Y-Plus-Out)
    req_ar_south   : out std_logic;               -- Request-Signal fr Arbiter-Logik (South-Out = Y-Minus-Out)
    req_ar_west    : out std_logic;               -- Request-Signal fr Arbiter-Logik (West-Out = X-Minus-Out)    
     -- Eingnge mux_ack_ar
    ack_ar_east    : in  nocdl_ctrl_type;         -- Acknowledge von Arbiter-Logik (East-Arbiter)
    ack_ar_north   : in  nocdl_ctrl_type;         -- Acknowledge von Arbiter-Logik (North-Arbiter)
    ack_ar_south   : in  nocdl_ctrl_type;         -- Acknowledge von Arbiter-Logik (South-Arbiter)
    ack_ar_west    : in  nocdl_ctrl_type;         -- Acknowledge von Arbiter-Logik (West-Arbiter)
     -- Eingnge mux_data
    data_in_east   : in  nocph_data_type;         -- Daten+Request vom East-Port
    data_in_north  : in  nocph_data_type;         -- Daten+Request vom North-Port
    data_in_south  : in  nocph_data_type;         -- Daten+Request vom South-Port
    data_in_west   : in  nocph_data_type;         -- Daten+Request vom West-Port
     -- Eingnge mux_ack
    ack_tx_east    : in  nocph_ctrl_type;         -- Acknowledge vom East-Port
    ack_tx_north   : in  nocph_ctrl_type;         -- Acknowledge vom North-Port
    ack_tx_south   : in  nocph_ctrl_type;         -- Acknowledge vom South-Port
    ack_tx_west    : in  nocph_ctrl_type);        -- Acknowledge vom West-Port
  
end noc_resport2d;

architecture noc_resport2d_arch of noc_resport2d is

  type state_type is (idle, wait_ar, wait_eof);

  -- Registertyp fr 2-Prozess-Methode
  type reg_type is record
                     state    : state_type;
                     free     : std_logic;          -- Port belegt oder nicht
                     req_ar   : std_logic;          -- Request-Signal fr Routing-Logik
                     switch   : std_logic_vector (2 downto 0);  -- Schalterstellung Demuxe
                     state_in : std_logic_vector (2 downto 0);  -- Schalter fr mux_ack
                   end record;

  signal r, rin   : reg_type;
  signal data_ack : nocph_ctrl_type;
  signal req_ar   : std_logic;
  signal ack_ar   : nocdl_ctrl_type;
  signal switch   : std_logic_vector (2 downto 0);
  signal state_in : std_logic_vector (2 downto 0);
  signal do_mux   : nocph_data_type;

  -- Konstanten fr die Schalter der (De)Multiplexer fr bessere Lesbarkeit im Code
  -- Definition erfolgt lokal, da unterschiedliche Gre in einigen Ports
  -- SWITCH schaltet die Multiplexer zu den Arbitern
  constant INIT            : std_logic_vector(2 downto 0) := "000";
  constant SWITCH_WEST     : std_logic_vector(2 downto 0) := "001";
  constant SWITCH_EAST     : std_logic_vector(2 downto 0) := "010";
  constant SWITCH_NORTH    : std_logic_vector(2 downto 0) := "011";
  constant SWITCH_SOUTH    : std_logic_vector(2 downto 0) := "100";
  constant SWITCH_RES      : std_logic_vector(2 downto 0) := "101";
  -- STATE_IN schaltet den Eingang an einen Ausgang
  constant STATE_IN_WEST   : std_logic_vector(2 downto 0) := "001";
  constant STATE_IN_EAST   : std_logic_vector(2 downto 0) := "010";
  constant STATE_IN_NORTH  : std_logic_vector(2 downto 0) := "011";
  constant STATE_IN_SOUTH  : std_logic_vector(2 downto 0) := "100";
  constant STATE_IN_RES    : std_logic_vector(2 downto 0) := "101";
  -- STATE_OUT schaltet den Ausgang an einen Eingang
  constant STATE_OUT_EAST  : std_logic_vector(2 downto 0) := "010";
  constant STATE_OUT_WEST  : std_logic_vector(2 downto 0) := "001";
  constant STATE_OUT_NORTH : std_logic_vector(2 downto 0) := "011";
  constant STATE_OUT_SOUTH : std_logic_vector(2 downto 0) := "100";
  constant STATE_OUT_RES   : std_logic_vector(2 downto 0) := "101";

  procedure multiplex_mux_ack (
    signal ack_tx_east  : in  nocph_ctrl_type;
    signal ack_tx_north : in  nocph_ctrl_type;
    signal ack_tx_south : in  nocph_ctrl_type;
    signal ack_tx_west  : in  nocph_ctrl_type;
    signal data_ack     : out nocph_ctrl_type;
    signal switch       : in  std_logic_vector (2 downto 0)) is
  begin
    case switch is
     when STATE_IN_EAST  => data_ack     <= ack_tx_east;
     when STATE_IN_NORTH => data_ack     <= ack_tx_north;
     when STATE_IN_SOUTH => data_ack     <= ack_tx_south;
     when STATE_IN_WEST  => data_ack     <= ack_tx_west;
     when others         => data_ack.ack <= '0';
    end case;
  end;

  procedure multiplex_mux_data (
    signal data_in_east  : in  nocph_data_type;
    signal data_in_north : in  nocph_data_type;
    signal data_in_south : in  nocph_data_type;
    signal data_in_west  : in  nocph_data_type;
    signal data_out      : out nocph_data_type;
    signal switch        : in  std_logic_vector (2 downto 0)) is
  begin
    case switch is
      when STATE_OUT_EAST  => data_out <= data_in_east;
      when STATE_OUT_NORTH => data_out <= data_in_north;
      when STATE_OUT_SOUTH => data_out <= data_in_south;
      when STATE_OUT_WEST  => data_out <= data_in_west;
      when others          => data_out <= nocph_data_type_default;
    end case;
  end;

  procedure demultiplex_demux_req(
    signal req_in        : in  std_logic;
    signal switch        : in  std_logic_vector(2 downto 0);
    signal req_out_east  : out std_logic;
    signal req_out_north : out std_logic;
    signal req_out_south : out std_logic;
    signal req_out_west  : out std_logic) is
  begin
    req_out_east  <= '0';
    req_out_north <= '0';
    req_out_south <= '0';
    req_out_west  <= '0';
    case switch is 
     when SWITCH_EAST  => req_out_east  <= req_in;
     when SWITCH_NORTH => req_out_north <= req_in;
     when SWITCH_SOUTH => req_out_south <= req_in;
     when SWITCH_WEST  => req_out_west  <= req_in;
     when others       => NULL;
    end case;
  end;
  
  procedure multiplex_mux_ack_ar (
    signal ack_ar_east  : in  nocdl_ctrl_type;
    signal ack_ar_north : in  nocdl_ctrl_type;
    signal ack_ar_south : in  nocdl_ctrl_type;
    signal ack_ar_west  : in  nocdl_ctrl_type;
    signal ack_ar_out   : out nocdl_ctrl_type;
    signal switch       : in  std_logic_vector (2 downto 0)) is
  begin
    case switch is
     when SWITCH_EAST  => ack_ar_out <= ack_ar_east;
     when SWITCH_NORTH => ack_ar_out <= ack_ar_north;
     when SWITCH_SOUTH => ack_ar_out <= ack_ar_south;
     when SWITCH_WEST  => ack_ar_out <= ack_ar_west;
     when others       => ack_ar_out.ack <= '0';
    end case;
  end;
  
begin  -- noc_resport2d_arch
  
    fifo : nocdl_fifo
      port map (
        clk        => clk,
        rst        => rst,
        write_data => do_mux,
        write_ctrl => ack_tx_intern,
        read_data  => data_out,
        read_ctrl  => ack_tx);

  -- port-logic to receive, analyze and switch frames
  logic: process (r, data_in, ack_ar.ack, data_ack.ack, state_out)
     variable v : reg_type;
  begin  -- process logic
    
    v := r;

    case r.state is
      when idle      =>  -- wenn ein neues Request vorliegt wird ein Header erwartet
                        if data_in.req = '1' and data_in.sof = '1' then
                          v.switch := get_target(SELF_ADRESS,data_in.data(noc_dstpart)); -- ermittle ZielOutput
                          v.req_ar   := '1';                   -- erzeuge ein Arbiter-Request, da neuer Header
                          v.free     := '0';                   -- Port ist wieder belegt
                          v.state    := wait_ar;
                        end if;
      when wait_ar   =>  -- warten auf das Arbiter-Acknowledge
                        if ack_ar.ack = '1' then
                          v.req_ar   := '0';                   -- Routing-Request ist wieder '0'        
                          v.state_in := r.switch;              -- verbinde den Eingang mit dem Ausgang
                          v.switch   := INIT;                  -- die Demuxe schalten nichts zum Arbiter
                          v.state    := wait_eof;
                        end if;
      when wait_eof  =>  -- warten auf das eof und das passende data_ack
                        if data_in.eof = '1' and data_ack.ack = '1' then
                          v.free       := '1';
                          v.state_in   := INIT;
                          v.state      := idle;
                        end if;
      when others    => v.state    := idle;
                        v.free     := '1';
                        v.state_in := INIT;
    end case;

    
    -- update registers
    rin <= v;

    -- drive outputs
    if r.free = '0' then
      ack_rx     <= data_ack;
    else
      ack_rx.ack <= '0';
    end if;
    data_in_intern <= data_in;          -- treibe Daten-Signal
    req_ar         <= r.req_ar;         -- treibe Arbiter-Request-Signal
    switch         <= r.switch;         -- treibe Schalter-Signal fr Demuxe
    free           <= r.free;           -- treibe Free-Signal
    state_in       <= r.state_in;       -- treibe Schalter fr Muxe
    
  end process logic;

  regs: process (clk)
  begin  -- process regs
    if rising_edge(clk) then
      -- update registers
      r <= rin;
      -- synch reset
      if rst = '0' then
        r.state    <= idle;
        r.free     <= '1';                    -- Port ist frei
        r.req_ar   <= '0';                    -- kein Request zur Routing-Logik
        r.switch   <= INIT;                   -- Demuxe schalten nichts weiter
        r.state_in <= INIT;                   -- mux_ack schaltet nichts weiter
      end if;
    end if;
  end process regs;  
      
  -- purpose: Multiplexer fr Acknowledge von anderen Ports
  mux_ack: process (state_in, ack_tx_east.ack, ack_tx_north.ack, ack_tx_south.ack, ack_tx_west.ack)
  begin  -- process mux_ack
    multiplex_mux_ack(ack_tx_east, ack_tx_north, ack_tx_south, ack_tx_west, data_ack, state_in);  
  end process mux_ack;
  
  -- purpose: Multiplexer fr Daten von anderen Ports
  mux_data : process (state_out, data_in_east, data_in_north, data_in_south, data_in_west)
  begin  -- process
    multiplex_mux_data(data_in_east, data_in_north, data_in_south, data_in_west, do_mux, state_out);  
  end process mux_data;

  -- purpose: Demultiplexer fr req zum richtigen Arbiter
  demux_req : process (switch, req_ar)
  begin  -- process demuxe
    demultiplex_demux_req(req_ar, switch, req_ar_east, req_ar_north, req_ar_south, req_ar_west);  
  end process demux_req;

  -- purpose: Mulitplexer fr ACK_AR vom Arbiter
  mux_ack_ar : process (switch, ack_ar_east.ack, ack_ar_north.ack, ack_ar_south.ack, ack_ar_west.ack)
  begin  -- process
    multiplex_mux_ack_ar(ack_ar_east, ack_ar_north, ack_ar_south, ack_ar_west, ack_ar, switch);
  end process mux_ack_ar;

end noc_resport2d_arch;
