-------------------------------------------------------------------------------
-- Description: Top-Entity of a 2D NoC Router to compose simple NxM mesh NoCs.
--              The Router consists of 5 Ports (North, East, West, South,
--              Resource) and implements the following features:
--              - XY Routing with absolute Adresses
--              - Wormhole Switching with FlowControl (req/ack) per Flit
--              - a single Flit-Buffer per Output Port
--              - RoundRobin Arbitration to schedule multiple Requests
--              - NO virtual channels, NO buffer FIFOs, NO traffic class
--                differentiation
--              - one global clock
--              - the router is device-independent and feasible for ASIC as
--                well as FPGA synthesis.
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use work.noc_pack.all;                  -- contains NoC-related parameters and types

entity noc_router2d is

  generic (
      self_adress   : std_logic_vector (3 downto 0) := "0101");
  
  port (
      clk           : in  std_logic;        -- Clock
      rst           : in  std_logic;        -- Reset
      -- south interface
      south_tx_data : out nocph_data_type;  -- Data Output to South
      south_tx_ctrl : in  nocph_ctrl_type;  
      south_rx_data : in  nocph_data_type;  -- Eingang am yPlus-Port 
      south_rx_ctrl : out nocph_ctrl_type;  
      -- north interface
      north_tx_data : out nocph_data_type;  -- Data Output to North
      north_tx_ctrl : in  nocph_ctrl_type;  
      north_rx_data : in  nocph_data_type;  -- Data Input from North
      north_rx_ctrl : out nocph_ctrl_type;  
      -- west interface
      west_tx_data  : out nocph_data_type;  -- Data Output to West
      west_tx_ctrl  : in  nocph_ctrl_type;  
      west_rx_data  : in  nocph_data_type;  -- Data Input from West
      west_rx_ctrl  : out nocph_ctrl_type;  
      -- east interface
      east_tx_data  : out nocph_data_type;  -- Data Output to East
      east_tx_ctrl  : in  nocph_ctrl_type;  
      east_rx_data  : in  nocph_data_type;  -- Data Input from East
      east_rx_ctrl  : out nocph_ctrl_type;  
      -- local port to resource
      res_tx_data   : out nocph_data_type;  -- Data Output to Resource
      res_tx_ctrl   : in  nocph_ctrl_type;  
      res_rx_data   : in  nocph_data_type;  -- Data Input from Resource
      res_rx_ctrl   : out nocph_ctrl_type); 
      
end noc_router2d;

architecture noc_router2d_arch of noc_router2d is

  -- Registers for West-Arbiter
  type reg_type_westar is record
                           ack_ar_east   : std_logic;     
                           ack_ar_res    : std_logic;
                           state_out     : std_logic_vector (1 downto 0);
                           free          : std_logic;
                           rr_reg        : std_logic; -- Toggle-Register for RoundRobin
                      end record;
  -- Registers for East-Arbiter
  type reg_type_eastar is record
                           ack_ar_west   : std_logic;     
                           ack_ar_res    : std_logic;
                           state_out     : std_logic_vector (1 downto 0);
                           free          : std_logic;
                           rr_reg        : std_logic; -- Toggle-Register for RoundRobin
                      end record;
  -- Registers for North-Arbiter
  type reg_type_northar is record
                           ack_ar_east   : std_logic;
                           ack_ar_west   : std_logic; 
                           ack_ar_south  : std_logic;     
                           ack_ar_res    : std_logic;
                           state_out     : std_logic_vector (2 downto 0);
                           free          : std_logic;
                           rr_reg        : integer range 0 to 3;  -- RoundRobin-Register
                       end record;
  -- Registers for South-Arbiter
  type reg_type_southar is record
                           ack_ar_east   : std_logic;
                           ack_ar_west   : std_logic; 
                           ack_ar_north  : std_logic;     
                           ack_ar_res    : std_logic;
                           state_out     : std_logic_vector (2 downto 0);
                           free          : std_logic;
                           rr_reg        : integer range 0 to 3;  -- RoundRobin-Register
                       end record;
  -- Registers for Res-Arbiter
  type reg_type_resar is record
                           ack_ar_east   : std_logic;
                           ack_ar_west   : std_logic; 
                           ack_ar_north  : std_logic;     
                           ack_ar_south  : std_logic;
                           state_out     : std_logic_vector (2 downto 0);
                           free          : std_logic;
                           rr_reg        : integer range 0 to 3;  -- RoundRobin-Register
                     end record;
  
  -- All Registers
  type reg_type is record
                           westar        : reg_type_westar;
                           eastar        : reg_type_eastar;
                           northar       : reg_type_northar;
                           southar       : reg_type_southar;
                           resar         : reg_type_resar;
                   end record;
  signal r, rin : reg_type;  -- Signals for 2-Procss-Coding-Style

  signal req_ar_easttowest,   req_ar_easttonorth  : std_logic;
  signal req_ar_easttosouth,  req_ar_easttores    : std_logic;
  signal req_ar_westtoeast,   req_ar_westtonorth  : std_logic;
  signal req_ar_westtosouth,  req_ar_westtores    : std_logic;
  signal req_ar_northtosouth, req_ar_northtores   : std_logic;
  signal req_ar_southtonorth, req_ar_southtores   : std_logic;
  signal req_ar_restowest,    req_ar_restonorth   : std_logic;
  signal req_ar_restosouth,   req_ar_restoeast    : std_logic;

  signal ack_ar_easttowest,   ack_ar_easttores    : nocdl_ctrl_type;
  signal ack_ar_westtoeast,   ack_ar_westtores    : nocdl_ctrl_type;
  signal ack_ar_northtowest,  ack_ar_northtoeast  : nocdl_ctrl_type;
  signal ack_ar_northtosouth, ack_ar_northtores   : nocdl_ctrl_type;
  signal ack_ar_southtowest,  ack_ar_southtonorth : nocdl_ctrl_type;
  signal ack_ar_southtoeast,  ack_ar_southtores   : nocdl_ctrl_type;
  signal ack_ar_restowest,    ack_ar_restonorth   : nocdl_ctrl_type;
  signal ack_ar_restosouth,   ack_ar_restoeast    : nocdl_ctrl_type;  

  signal ack_tx_west, ack_tx_north, ack_tx_south   : nocph_ctrl_type;
  signal ack_tx_east, ack_tx_res                   : nocph_ctrl_type;
  signal data_in_east, data_in_west, data_in_north : nocph_data_type;
  signal data_in_res, data_in_south                : nocph_data_type;
  
  signal state_out_north, state_out_south, state_out_res          : std_logic_vector (2 downto 0);
  signal state_out_west, state_out_east                           : std_logic_vector (1 downto 0);
  signal req_vector_westar, req_vector_eastar                     : std_logic_vector (1 downto 0);
  signal req_vector_northar, req_vector_southar, req_vector_resar : std_logic_vector (3 downto 0);
  signal free_west, free_east, free_res, free_north, free_south   : std_logic;

  -- Constants for (De)Multiplexer (for better Readability)
  constant EAST2    : std_logic_vector(1 downto 0) := "10";
  constant WEST2    : std_logic_vector(1 downto 0) := "11";
  constant RES2     : std_logic_vector(1 downto 0) := "01";
  constant EAST     : std_logic_vector(2 downto 0) := "010";
  constant WEST     : std_logic_vector(2 downto 0) := "001";
  constant RES      : std_logic_vector(2 downto 0) := "101";
  constant NORTH    : std_logic_vector(2 downto 0) := "011";
  constant SOUTH    : std_logic_vector(2 downto 0) := "100";
  constant NOREQ    : std_logic_vector(3 downto 0) := "0000";

  -- Macro Function (returns the number of request in an request-vector)
  function no_of_ones (signal vector : std_logic_vector(3 downto 0))
    return natural is    
  begin  -- no_of_ones
    case vector is
	   when "0000" => return 0;
		when "0001" => return 1;
		when "0010" => return 1;
		when "0011" => return 2;
		when "0100" => return 1;
		when "0101" => return 2;
		when "0110" => return 2;
		when "0111" => return 3;
		when "1000" => return 1;
		when "1001" => return 2;
		when "1010" => return 2;
		when "1011" => return 3;
		when "1100" => return 2;
		when "1101" => return 3;
		when "1110" => return 3;
		when "1111" => return 4;
                when others => return 0;
	 end case;
  end no_of_ones;
    

begin  -- noc_router2d_arch

  -- Instances of the ports

  west_port : noc_westport2d
     generic map (
       self_adress    => self_adress)
     port map (
       clk            => clk,
       rst            => rst,
       data_in        => west_rx_data,
       data_out       => west_tx_data,
       ack_rx         => west_rx_ctrl,
       ack_tx         => west_tx_ctrl,
       free           => free_west,
       data_in_intern => data_in_west,
       ack_tx_intern  => ack_tx_west,
       state_out      => state_out_west,
       req_ar_east    => req_ar_westtoeast,
       req_ar_north   => req_ar_westtonorth,
       req_ar_south   => req_ar_westtosouth,
       req_ar_res     => req_ar_westtores,
       ack_ar_east    => ack_ar_easttowest,
       ack_ar_north   => ack_ar_northtowest,
       ack_ar_south   => ack_ar_southtowest,
       ack_ar_res     => ack_ar_restowest,
       data_in_east   => data_in_east,
       data_in_res    => data_in_res,
       ack_tx_east    => ack_tx_east,
       ack_tx_north   => ack_tx_north,
       ack_tx_south   => ack_tx_south,
       ack_tx_res     => ack_tx_res);
    
  east_port : noc_eastport2d
    generic map (
       self_adress    => self_adress)
     port map (
       clk            => clk,
       rst            => rst,
       data_in        => east_rx_data,
       data_out       => east_tx_data,
       ack_rx         => east_rx_ctrl,
       ack_tx         => east_tx_ctrl,
       free           => free_east,
       data_in_intern => data_in_east,
       ack_tx_intern  => ack_tx_east,
       state_out      => state_out_east,
       req_ar_west    => req_ar_easttowest,
       req_ar_north   => req_ar_easttonorth,
       req_ar_south   => req_ar_easttosouth,
       req_ar_res     => req_ar_easttores,
       ack_ar_west    => ack_ar_westtoeast,
       ack_ar_north   => ack_ar_northtoeast,
       ack_ar_south   => ack_ar_southtoeast,
       ack_ar_res     => ack_ar_restoeast,
       data_in_west   => data_in_west,
       data_in_res    => data_in_res,
       ack_tx_west    => ack_tx_west,
       ack_tx_north   => ack_tx_north,
       ack_tx_south   => ack_tx_south,
       ack_tx_res     => ack_tx_res);
    
  north_port : noc_northport2d
    generic map (
       self_adress    => self_adress)
     port map (
       clk            => clk,
       rst            => rst,
       data_in        => north_rx_data,
       data_out       => north_tx_data,
       ack_rx         => north_rx_ctrl,
       ack_tx         => north_tx_ctrl,
       free           => free_north,
       data_in_intern => data_in_north,
       ack_tx_intern  => ack_tx_north,
       state_out      => state_out_north,
       req_ar_south   => req_ar_northtosouth,
       req_ar_res     => req_ar_northtores,
       ack_ar_south   => ack_ar_southtonorth,
       ack_ar_res     => ack_ar_restonorth,
       data_in_west   => data_in_west,
       data_in_east   => data_in_east,
       data_in_south  => data_in_south,
       data_in_res    => data_in_res,
       ack_tx_south   => ack_tx_south,
       ack_tx_res     => ack_tx_res);
    
  south_port : noc_southport2d
    generic map (
       self_adress    => self_adress)
     port map (
       clk            => clk,
       rst            => rst,
       data_in        => south_rx_data,
       data_out       => south_tx_data,
       ack_rx         => south_rx_ctrl,
       ack_tx         => south_tx_ctrl,
       free           => free_south,
       data_in_intern => data_in_south,
       ack_tx_intern  => ack_tx_south,
       state_out      => state_out_south,
       req_ar_north   => req_ar_southtonorth,
       req_ar_res     => req_ar_southtores,
       ack_ar_north   => ack_ar_northtosouth,
       ack_ar_res     => ack_ar_restosouth,
       data_in_west   => data_in_west,
       data_in_east   => data_in_east,
       data_in_north  => data_in_north,
       data_in_res    => data_in_res,
       ack_tx_north   => ack_tx_north,
       ack_tx_res     => ack_tx_res);

  res_port : noc_resport2d
    generic map (
       self_adress    => self_adress)
     port map (
       clk            => clk,
       rst            => rst,
       data_in        => res_rx_data,
       data_out       => res_tx_data,
       ack_rx         => res_rx_ctrl,
       ack_tx         => res_tx_ctrl,
       free           => free_res,
       data_in_intern => data_in_res,
       ack_tx_intern  => ack_tx_res,
       state_out      => state_out_res,
       req_ar_west    => req_ar_restowest,
       req_ar_east    => req_ar_restoeast,
       req_ar_north   => req_ar_restonorth,
       req_ar_south   => req_ar_restosouth,
       ack_ar_west    => ack_ar_westtores,
       ack_ar_east    => ack_ar_easttores,
       ack_ar_north   => ack_ar_northtores,
       ack_ar_south   => ack_ar_southtores,
       data_in_west   => data_in_west,
       data_in_east   => data_in_east,
       data_in_north  => data_in_north,
       data_in_south  => data_in_south,
       ack_tx_west    => ack_tx_west,
       ack_tx_east    => ack_tx_east,
       ack_tx_north   => ack_tx_north,
       ack_tx_south   => ack_tx_south);

  -- Request-Vectors for the Arbiter
  req_vector_westar(0)  <= req_ar_easttowest;
  req_vector_westar(1)  <= req_ar_restowest;
  req_vector_eastar(0)  <= req_ar_westtoeast;
  req_vector_eastar(1)  <= req_ar_restoeast;
  req_vector_northar(0) <= req_ar_westtonorth;
  req_vector_northar(1) <= req_ar_easttonorth;
  req_vector_northar(2) <= req_ar_southtonorth;
  req_vector_northar(3) <= req_ar_restonorth;
  req_vector_southar(0) <= req_ar_westtosouth;
  req_vector_southar(1) <= req_ar_easttosouth;
  req_vector_southar(2) <= req_ar_northtosouth;
  req_vector_southar(3) <= req_ar_restosouth;
  req_vector_resar(0)   <= req_ar_westtores;
  req_vector_resar(1)   <= req_ar_easttores;
  req_vector_resar(2)   <= req_ar_northtores;
  req_vector_resar(3)   <= req_ar_southtores;

  -- in        : r.westar, req_vector_westar, free_res, free_east
  -- out       : ack_ar_westtoeast, ack_ar_westtores state_out_west
  -- purpose   : Arbiter for West-Output
  west_arbiter : process (r.westar, req_vector_westar, free_res, free_east)
    variable v : reg_type_westar;  -- local variable for 2-Process-Coding-Style
  begin

    -- load old Values
    v := r.westar;
    
    -- default assignments
    v.ack_ar_east := '0';     
    v.ack_ar_res  := '0';
    
    -- update port-state after transmission of a frame from another port is complete
    if r.westar.state_out = EAST2 and free_east = '1' and r.westar.free = '0' then
      v.free      := free_east;
      v.state_out := (others => '0');
    end if;
    if r.westar.state_out = RES2 and free_res = '1' and r.westar.free = '0' then
      v.free      := free_res;
      v.state_out := (others => '0');
    end if;
    -- only when idle, a new frame can be accepted
    if r.westar.free = '1' then      
      -- analyze pending requests and return acknowledge to the scheduled port
      case req_vector_westar is                     
        when "01" => grant_request(EAST2, v.state_out, v.free, v.ack_ar_east);
        when "10" => grant_request(RES2, v.state_out, v.free, v.ack_ar_res);
        when "11" => case r.westar.rr_reg is
                       when '0' => grant_request(EAST2, v.state_out, v.free, v.ack_ar_east);
                       when '1' => grant_request(RES2, v.state_out, v.free, v.ack_ar_res);
                       when others => null;
                     end case;
                     v.rr_reg := not r.westar.rr_reg;  -- toggle RoundRobin-Register
        when others => NULL;
      end case;
    end if;
    
    -- update Registers
    rin.westar <= v;
    
    -- drive Outputs
    ack_ar_westtoeast.ack <= r.westar.ack_ar_east;  
    ack_ar_westtores.ack  <= r.westar.ack_ar_res;   
    state_out_west        <= r.westar.state_out;
    
  end process west_arbiter;
  

  -- in        : r.eastar, req_vector_eastar, free_res, free_west
  -- out       : ack_ar_easttowest, ack_ar_easttores state_out_east
  -- purpose   : Arbiter for East Output
  east_arbiter : process (r.eastar, req_vector_eastar, free_res, free_west)
    variable v : reg_type_eastar;  -- local variable for 2-Process-Coding-Style
  begin
    
    -- load old Values
    v := r.eastar;
    
    -- default assignments
    v.ack_ar_west  := '0';     
    v.ack_ar_res   := '0';
    
    -- update port-state after transmission of a frame from another port is complete
    if r.eastar.state_out = WEST2 and free_west = '1' and r.eastar.free = '0' then
      v.free      := free_west;
      v.state_out := (others => '0');
    end if;
    if r.eastar.state_out = RES2 and free_res = '1' and r.eastar.free = '0' then
      v.free      := free_res;
      v.state_out := (others => '0');
    end if;
    -- only when idle, a new frame can be accepted
    if r.eastar.free = '1' then      
      -- analyze pending requests
      case req_vector_eastar is                     
        when "01" => grant_request(WEST2, v.state_out, v.free, v.ack_ar_west);        
        when "10" => grant_request(RES2, v.state_out, v.free, v.ack_ar_res);
        when "11" => case r.eastar.rr_reg is
                       when '0' => grant_request(WEST2, v.state_out, v.free, v.ack_ar_west);
                       when '1' => grant_request(RES2, v.state_out, v.free, v.ack_ar_res);
                       when others => null;
                     end case;
                     v.rr_reg := not r.eastar.rr_reg;  -- toggle RoundRobin-Register
       when others => NULL;
      end case;
    end if;
    
    -- update Registers
    rin.eastar <= v;
    
    -- drive Outputs
    ack_ar_easttowest.ack <= r.eastar.ack_ar_west;  
    ack_ar_easttores.ack  <= r.eastar.ack_ar_res;   
    state_out_east        <= r.eastar.state_out;
    
  end process east_arbiter;
  
  
  -- in         : r.northar, req_vector_northar, free_res, free_west, free_south, free_east
  -- out        : ack_ar_northtowest, ack_ar_northtores, ack_ar_northtosouth, ack_ar_northtoeast, state_out_north
  -- purpose    : Arbiter for North Output
  north_arbiter : process (r.northar, req_vector_northar, free_res, free_east, free_south, free_west)
    variable v  : reg_type_northar;  -- local variable for 2-Process-Coding-Style
  begin
    
    -- load old Values
    v := r.northar;
    
    -- default assignments
    v.ack_ar_west  := '0';     
    v.ack_ar_east  := '0';
    v.ack_ar_south := '0';
    v.ack_ar_res   := '0';

    -- update port-state after transmission of a frame from another port is complete
    if r.northar.state_out = WEST and free_west = '1' and r.northar.free = '0' then
      v.free      := free_west;
      v.state_out := (others => '0');
    end if;
    if r.northar.state_out = EAST and free_east = '1' and r.northar.free = '0' then
      v.free      := free_east;
      v.state_out := (others => '0');
    end if;
    if r.northar.state_out = SOUTH and free_south = '1' and r.northar.free = '0' then
      v.free      := free_south;
      v.state_out := (others => '0');
    end if;
    if r.northar.state_out = RES and free_res = '1' and r.northar.free = '0' then
      v.free      := free_res;
      v.state_out := (others => '0');
    end if;
    -- only when idle, a new frame can be accepted
    if r.northar.free = '1' and req_vector_northar /= NOREQ then
      -- analyze pending requests
      for i in 0 to 2 loop
        if req_vector_northar(v.rr_reg) = '0' then
          if v.rr_reg = 3 then
            v.rr_reg := 0;
          else
            v.rr_reg := v.rr_reg + 1;
          end if;
        end if;
      end loop;
      case v.rr_reg is
        when 0      => grant_request5ports(WEST, v.state_out, v.free, v.ack_ar_west);
        when 1      => grant_request5ports(EAST, v.state_out, v.free, v.ack_ar_east);
        when 2      => grant_request5ports(SOUTH, v.state_out, v.free, v.ack_ar_south);
        when 3      => grant_request5ports(RES, v.state_out, v.free, v.ack_ar_res);
        when others => null;
      end case;
      if no_of_ones(req_vector_northar) > 1 then
        if r.northar.rr_reg = 3 then
          v.rr_reg := 0;
        else
          v.rr_reg := r.northar.rr_reg + 1;
        end if;
      else
        v.rr_reg := r.northar.rr_reg;
      end if;
    end if;
    
    -- update Registers
    rin.northar <= v;
    
    -- drive Outputs
    ack_ar_northtowest.ack  <= r.northar.ack_ar_west;  
    ack_ar_northtoeast.ack  <= r.northar.ack_ar_east;  
    ack_ar_northtosouth.ack <= r.northar.ack_ar_south; 
    ack_ar_northtores.ack   <= r.northar.ack_ar_res;   
    state_out_north         <= r.northar.state_out;    
  end process north_arbiter;
  
  
  -- in         : r.southar, req_vector_southar, free_res, free_west, free_north, free_east
  -- out        : ack_ar_southtowest, ack_ar_southtores, ack_ar_southtonorth, ack_ar_southtoeast, state_out_south
  -- purpose    : Arbiter for South Output
  south_arbiter : process (r.southar, req_vector_southar, free_res, free_east, free_north, free_west)
    variable v  : reg_type_southar; -- local variable for 2-Process-Coding-Style
  begin
    
    -- load old Values
    v := r.southar;
    
    -- Defaultwerte fr Acknowledge-Signale
    v.ack_ar_west  := '0';     
    v.ack_ar_east  := '0';
    v.ack_ar_north := '0';
    v.ack_ar_res   := '0';

    -- update port-state after transmission of a frame from another port is complete
    if r.southar.state_out = WEST and free_west = '1' and r.southar.free = '0' then
      v.free      := free_west;
      v.state_out := (others => '0');
    end if;
    if r.southar.state_out = EAST and free_east = '1' and r.southar.free = '0' then
      v.free      := free_east;
      v.state_out := (others => '0');
    end if;
    if r.southar.state_out = NORTH and free_north = '1' and r.southar.free = '0' then
      v.free      := free_north;
      v.state_out := (others => '0');
    end if;
    if r.southar.state_out = RES and free_res = '1' and r.southar.free = '0' then
      v.free      := free_res;
      v.state_out := (others => '0');
    end if;
    -- only when idle, a new frame can be accepted
    if r.southar.free = '1' and req_vector_southar /= NOREQ then
      -- analyze pending requests
      for i in 0 to 2 loop
        if req_vector_southar(v.rr_reg) = '0' then
          if v.rr_reg = 3 then
            v.rr_reg := 0;
          else
            v.rr_reg := v.rr_reg + 1;
          end if;
        end if;
      end loop;
      case v.rr_reg is
        when 0      => grant_request5ports(WEST, v.state_out, v.free, v.ack_ar_west);
        when 1      => grant_request5ports(EAST, v.state_out, v.free, v.ack_ar_east);
        when 2      => grant_request5ports(NORTH, v.state_out, v.free, v.ack_ar_north);
        when 3      => grant_request5ports(RES, v.state_out, v.free, v.ack_ar_res);
        when others => null;
      end case;
      if no_of_ones(req_vector_southar) > 1 then
        if r.southar.rr_reg = 3 then
          v.rr_reg := 0;
        else
          v.rr_reg := r.southar.rr_reg + 1;
        end if;
      else
        v.rr_reg := r.southar.rr_reg;
      end if;
    end if;
    
    -- update Registers
    rin.southar <= v;
    
    -- drive Outputs
    ack_ar_southtowest.ack  <= r.southar.ack_ar_west;  
    ack_ar_southtoeast.ack  <= r.southar.ack_ar_east;  
    ack_ar_southtonorth.ack <= r.southar.ack_ar_north; 
    ack_ar_southtores.ack   <= r.southar.ack_ar_res;   
    state_out_south         <= r.southar.state_out;
    
  end process south_arbiter;
  

  -- in        : r.resar, req_vector_resar, free_south, free_west, free_north, free_east
  -- out       : ack_ar_restowest, ack_ar_restosouth, ack_ar_restonorth, ack_ar_restoeast, state_out_res
  -- purpose   : Arbiter for Resource Output
  res_arbiter  : process (r.resar, req_vector_resar, free_south, free_east, free_north, free_west)
    variable v : reg_type_resar;  -- local variable for 2-Process-Coding-Style
  begin

    -- load old Values
    v := r.resar; 

    -- default assignments
    v.ack_ar_west  := '0';     
    v.ack_ar_east  := '0';
    v.ack_ar_north := '0';
    v.ack_ar_south := '0';

    -- update port-state after transmission of a frame from another port is complete
    if r.resar.state_out = WEST and free_west = '1' and r.resar.free = '0' then
      v.free      := free_west;
      v.state_out := (others => '0');
    end if;
    if r.resar.state_out = EAST and free_east = '1' and r.resar.free = '0' then
      v.free      := free_east;
      v.state_out := (others => '0');
    end if;
    if r.resar.state_out = NORTH and free_north = '1' and r.resar.free = '0' then
      v.free      := free_north;
      v.state_out := (others => '0');
    end if;
    if r.resar.state_out = SOUTH and free_south = '1' and r.resar.free = '0' then
      v.free      := free_south;
      v.state_out := (others => '0');
    end if;
    -- only when idle, a new frame can be accepted
    if r.resar.free = '1' and req_vector_resar /= NOREQ then
      -- analyze pending requests
      for i in 0 to 2 loop
        if req_vector_resar(v.rr_reg) = '0' then
          if v.rr_reg = 3 then
            v.rr_reg := 0;
          else
            v.rr_reg := v.rr_reg + 1;
          end if;
        end if;
      end loop;
      case v.rr_reg is
        when 0      => grant_request5ports(WEST, v.state_out, v.free, v.ack_ar_west);
        when 1      => grant_request5ports(EAST, v.state_out, v.free, v.ack_ar_east);
        when 2      => grant_request5ports(NORTH, v.state_out, v.free, v.ack_ar_north);
        when 3      => grant_request5ports(SOUTH, v.state_out, v.free, v.ack_ar_south);
        when others => null;
      end case;
      if no_of_ones(req_vector_resar) > 1 then
        if r.resar.rr_reg = 3 then
          v.rr_reg := 0;
        else
          v.rr_reg := r.resar.rr_reg + 1;
        end if;
      else
        v.rr_reg := r.resar.rr_reg;
      end if;
    end if;
    
    -- update Registers
    rin.resar <= v;

    -- drive Outputs
    ack_ar_restowest.ack  <= r.resar.ack_ar_west;  
    ack_ar_restoeast.ack  <= r.resar.ack_ar_east;  
    ack_ar_restonorth.ack <= r.resar.ack_ar_north; 
    ack_ar_restosouth.ack <= r.resar.ack_ar_south;   
    state_out_res         <= r.resar.state_out;
    
  end process res_arbiter;

  -- clocked register process
  regs : process (clk, rst)
  begin
    if rising_edge(clk) then
      -- update registers
      r <= rin;
      -- synch reset
      -- define reset values for each register
      if rst = '0' then
        r.westar.free          <= '1';
        r.westar.state_out     <= (others => '0');
        r.westar.rr_reg        <= '0';
        r.westar.ack_ar_res    <= '0';
        r.westar.ack_ar_east   <= '0';
        r.eastar.free          <= '1';
        r.eastar.state_out     <= (others => '0');
        r.eastar.rr_reg        <= '0';
        r.eastar.ack_ar_west   <= '0';
        r.eastar.ack_ar_res    <= '0';
        r.northar.free         <= '1';
        r.northar.state_out    <= (others => '0');
        r.northar.rr_reg       <= 0;
        r.northar.ack_ar_west  <= '0';
        r.northar.ack_ar_east  <= '0';
        r.northar.ack_ar_res   <= '0';
        r.northar.ack_ar_south <= '0';
        r.southar.free         <= '1';
        r.southar.state_out    <= (others => '0');
        r.southar.rr_reg       <= 0;
        r.southar.ack_ar_west  <= '0';
        r.southar.ack_ar_east  <= '0';
        r.southar.ack_ar_north <= '0';
        r.southar.ack_ar_res   <= '0';
        r.resar.free           <= '1';
        r.resar.state_out      <= (others => '0');
        r.resar.rr_reg         <= 0;
        r.resar.ack_ar_west    <= '0';
        r.resar.ack_ar_east    <= '0';
        r.resar.ack_ar_north   <= '0';
        r.resar.ack_ar_south   <= '0';
      end if;
    end if; 
  end process regs;

end noc_router2d_arch;
