-------------------------------------------------------------------------------
-- Description: This FIFO is used to store Flits in a Wormhole-switched NoC
--              router. 
--              Currently, the capacity is reduced to one flit only.
--              You may insert other buffering mechanisms or memories here.
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.noc_pack.all;

entity nocdl_fifo is
  port (
    rst        : in  std_logic;
    clk        : in  std_logic;
    write_data : in  nocph_data_type;
    write_ctrl : out nocph_ctrl_type;
    read_data  : out nocph_data_type;
    read_ctrl  : in  nocph_ctrl_type);
end nocdl_fifo;


architecture rtl of nocdl_fifo is

  type state_type is (idle, transmit);

  type reg_type is
         record
           buff  : nocph_data_type;  -- the flitbuffer for 1 flit
           ack   : std_logic;
           state : state_type;
         end record;
  
  signal r, rin : reg_type;

begin  -- rtl
  
  regs : process (clk, rst)
  begin
    if rising_edge(clk) then
      -- update registers
      r <= rin;
      -- synch reset
      if rst = '0' then
        r.buff  <= nocph_data_type_default;
        r.ack   <= '0';
        r.state <= idle;
      end if;
    end if;
  end process regs;
 

  comb : process (r, write_data, read_ctrl)
    variable v : reg_type;
  begin

    -- default assignments
    v     := r;
    v.ack := '0';
    
    case r.state is
      when idle     => if write_data.req = '1' then
                         v.buff    := write_data;
                         v.ack     := '1';
                         v.state   := transmit;
                       else
                         v.state   := idle;
                       end if;
      when transmit => if read_ctrl.ack = '1' then
                         v.buff    := write_data;
                         if r.buff.eof = '1' then
                           v.state := idle;
                           v.buff  := nocph_data_type_default;
                         else
                           v.ack   := '1';
                           v.state := transmit;
                         end if;
                       end if;
      when others   => v.state     := idle;
    end case;
    
    -- update register
    rin <= v;
    
    -- drive outputs
    read_data      <= r.buff;
    write_ctrl.ack <= r.ack;
  
  end process comb;
  
end rtl;

